_MAJOR = "0"
_MINOR = "1"
# On main and in a nightly release the patch should be one ahead of the last
# released build.
_PATCH = "1"
# This is mainly for nightly builds which have the suffix ".dev$DATE". See
# https://semver.org/#is-v123-a-semantic-version for the semantics.
_SUFFIX = ""

VERSION_SHORT = "{0}.{1}".format(_MAJOR, _MINOR)
VERSION = "{0}.{1}.{2}{3}".format(_MAJOR, _MINOR, _PATCH, _SUFFIX)



from abc import abstractmethod
from dataclasses import dataclass, field
import getpass
import os
from pathlib import Path
import platform
import time
from typing import List, Tuple, Iterable, Any, Dict, Optional
import socket
import sys

import click

from tango.common.from_params import FromParams
from tango.common.params import Params
from tango.common.registrable import Registrable
from tango.step import Step
from tango.step_cache import StepCache
from tango.step_graph import StepGraph, StepStub


class Executor(Registrable):
    """
    An ``Executor`` is :class:`~tango.common.registrable.Registrable` class that is
    responsible for running steps and returning them with their result.

    Subclasses should implement :meth:`execute_step_group()`.
    """

    default_implementation = "simple"
    """
    The default implementation is :class:`SimpleExecutor`.
    """

    def __init__(self, include_package: Optional[List[str]] = None) -> None:
        self.include_package = include_package

    def execute_step_graph(self, step_graph: StepGraph, step_cache: StepCache) -> List[Step]:
        """
        Execute an entire :class:`tango.step_graph.StepGraph`.
        """
        # Keeps track of all steps that we've ran so far by name, and stores the results
        # for ones that can't be cached.
        executed: Dict[str, Tuple[Step, Any]] = {}
        remaining: List[StepStub] = list(step_graph)
        group: List[Step] = []
        out: List[Step] = []

        def run_group():
            # Gather all steps that need to be ran (no cache hit).
            needed: List[Step] = []
            for step in group:
                if step not in step_cache:
                    needed.append(step)
                else:
                    executed[step.name] = (step, None)
                    click.echo(
                        click.style("✓ Found output for ", fg="green")
                        + click.style(f'"{step.name}"', bold=True, fg="green")
                        + click.style(" in cache", fg="green")
                    )
            if needed:
                # Run the needed ones.
                for step, result in self.execute_step_group(needed, step_cache):
                    # Add to cache or keep result in `executed`.
                    if step.cache_results and step not in step_cache:
                        step_cache[step] = result
                        executed[step.name] = (step, None)
                    else:
                        executed[step.name] = (step, result)

        while remaining:
            next_step_ready = True
            for ref in remaining[0].dependencies:
                if ref not in executed:
                    # Still has dependencies that need to be ran first.
                    next_step_ready = False
                    break

            if not next_step_ready:
                # Run the current group.
                assert group
                run_group()
                out.extend(group)
                group = []

            # Materialize the next step.
            next_up = remaining.pop(0)
            config = self._replace_refs_with_results(next_up.config, executed, step_cache)
            step = Step.from_params(Params(config), step_name=next_up.name)
            step._executor = self
            group.append(step)

        # Finish up last group.
        if group:
            run_group()
            out.extend(group)
        return out

    @abstractmethod
    def execute_step_group(
        self, step_group: List[Step], step_cache: StepCache
    ) -> Iterable[Tuple[Step, Any]]:
        """
        Execute all steps in the group, returning them with their results.

        The executor can assume that all steps in the group are independent (none of them
        depend on the result of any other step in the group), so they can be ran
        in any order or even in parallel.
        """
        raise NotImplementedError

    @staticmethod
    def _replace_refs_with_results(
        o: Any, executed: Dict[str, Tuple[Step, Any]], step_cache: StepCache
    ) -> Any:
        if isinstance(o, list):
            return [Executor._replace_refs_with_results(x, executed, step_cache) for x in o]
        elif isinstance(o, tuple):
            return tuple(Executor._replace_refs_with_results(list(o), executed, step_cache))
        elif isinstance(o, set):
            return set(Executor._replace_refs_with_results(list(o), executed, step_cache))
        elif isinstance(o, dict):
            if set(o.keys()) == {"type", "ref"}:
                if o["ref"] in executed:
                    step, result = executed[o["ref"]]
                    if result is not None:
                        return result
                    else:
                        return step_cache[step]
                else:
                    raise ValueError(f"result for step '{o['ref']}' could not be found!")
            else:
                return {
                    k: Executor._replace_refs_with_results(v, executed, step_cache)
                    for k, v in o.items()
                }
        elif isinstance(o, (str, bool, int, float)):
            return o
        else:
            raise ValueError(o)


@Executor.register("simple")
class SimpleExecutor(Executor):
    """
    A simple :class:`Executor` that just runs all steps locally one at a time.

    .. tip::
        Registered as an :class:`Executor` under the name "simple".

    """

    def execute_step_group(
        self, step_group: List[Step], step_cache: StepCache
    ) -> Iterable[Tuple[Step, Any]]:
        for step in step_group:
            click.echo(
                click.style("● Starting run for ", fg="blue")
                + click.style(f'"{step.name}"', bold=True, fg="blue")
            )
            metadata = ExecutorMetadata(step=step.unique_id)
            result = step.result(step_cache)
            metadata.finalize()
            metadata.to_params().to_file(step_cache.path_for_step(step) / "executor-metadata.json")
            click.echo(
                click.style("✓ Finished run for ", fg="green")
                + click.style(f'"{step.name}"', bold=True, fg="green")
            )
            yield step, result


@dataclass
class PlatformMetadata(FromParams):
    python: str = platform.python_version()
    operating_system: str = platform.platform()
    executable: Path = Path(sys.executable)
    cpu_count: Optional[int] = os.cpu_count()
    user: str = getpass.getuser()
    host: str = socket.gethostname()
    root: Path = Path(os.getcwd())


@dataclass
class GitMetadata(FromParams):
    commit: Optional[str] = None
    remote: Optional[str] = None

    @classmethod
    def check_for_repo(cls) -> Optional["GitMetadata"]:
        import subprocess

        try:
            commit = (
                subprocess.check_output("git rev-parse HEAD".split(" ")).decode("ascii").strip()
            )
            remote: Optional[str] = None
            for line in (
                subprocess.check_output("git remote -v".split(" "))
                .decode("ascii")
                .strip()
                .split("\n")
            ):
                if "(fetch)" in line:
                    _, line = line.split("\t")
                    remote = line.split(" ")[0]
                    break
            return cls(commit=commit, remote=remote)
        except subprocess.CalledProcessError:
            return None


@dataclass
class ExecutorMetadata(FromParams):
    step: str
    platform: PlatformMetadata = PlatformMetadata()
    git: Optional[GitMetadata] = GitMetadata.check_for_repo()
    started_at: float = field(default_factory=time.time)
    finished_at: Optional[float] = None
    duration: Optional[float] = None

    def finalize(self):
        self.finished_at = time.time()
        self.duration = round(self.finished_at - self.started_at, 4)



from dataclasses import dataclass, field
from typing import Mapping, Any, Sequence, TypeVar, Generic


T = TypeVar("T")


@dataclass
class DatasetDict(Generic[T]):
    """
    This definition of a dataset combines all splits and arbitrary metadata into
    one handy class.
    """

    splits: Mapping[str, Sequence[T]]
    """
    Maps the name of the split to a sequence of instances of type ``T``.
    """

    metadata: Mapping[str, Any] = field(default_factory=dict)
    """Metadata can contain anything you need."""

    def __getitem__(self, split: str) -> Sequence[T]:
        """
        Get a split.
        """
        return self.splits[split]

    def __len__(self) -> int:
        """
        Returns the number of splits.
        """
        return len(self.splits)

    def keys(self):
        return self.splits.keys()



from contextlib import contextmanager
from copy import deepcopy
import logging
import os
import shutil
import tempfile
import typing as t
from pathlib import Path

from .registrable import Registrable
from .util import PathOrStr


class TangoTestCase:
    """
    A custom testing class that

    * disables some of the more verbose logging,
    * creates and destroys a temp directory as a test fixture, and
    * restores the internal state of the `Registrable` class at the end of each test method.

    """

    PROJECT_ROOT = (Path(__file__).parent / ".." / "..").resolve()
    """
    Root of the git repository.
    """

    MODULE_ROOT = PROJECT_ROOT / "tango"
    """
    Root of the tango module.
    """

    TESTS_ROOT = PROJECT_ROOT / "tests"
    """
    Root of the tests directory.
    """

    FIXTURES_ROOT = PROJECT_ROOT / "test_fixtures"
    """
    Root of the test fixtures directory.
    """

    def setup_method(self):
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s - %(message)s", level=logging.DEBUG
        )

        # Disabling some of the more verbose logging statements that typically aren't very helpful
        # in tests.
        logging.getLogger("urllib3.connectionpool").disabled = True

        # Create a temporary scratch directory.
        self.TEST_DIR = Path(tempfile.mkdtemp(prefix="tango_tests"))
        os.makedirs(self.TEST_DIR, exist_ok=True)

    @classmethod
    def setup_class(cls):
        # During teardown we'll restore the state of `Registrable`'s internal registry
        # to make sure any registered mock test classes are removed so they don't conflict
        # with other tests.
        cls._original_registry = deepcopy(Registrable._registry)

    def teardown_method(self):
        shutil.rmtree(self.TEST_DIR)

    @classmethod
    def teardown_class(cls):
        Registrable._registry = cls._original_registry

    def run(
        self,
        config: t.Union[PathOrStr, t.Dict[str, t.Any]],
        overrides: t.Optional[str] = None,
        include_package: t.Optional[t.List[str]] = None,
    ) -> Path:
        from .params import Params
        from tango.__main__ import _run

        if isinstance(config, dict):
            params = Params(config)
            config = self.TEST_DIR / "config.json"
            params.to_file(t.cast(Path, config))

        run_dir = self.TEST_DIR / "run"
        _run(
            str(config),
            directory=str(run_dir),
            overrides=overrides,
            include_package=include_package,
        )
        return run_dir


@contextmanager
def run_experiment(config: t.Union[PathOrStr, t.Dict[str, t.Any]]):
    """
    A context manager to make testing experiments easier. On ``__enter__`` it runs
    the experiment and returns the path to the cache directory, a temporary directory that will be
    cleaned up on ``__exit__``.
    """
    test_case = TangoTestCase()
    try:
        test_case.setup_method()
        yield test_case.run(config)
    finally:
        test_case.teardown_method()



import os
import warnings

from filelock import FileLock as _FileLock, Timeout, AcquireReturnProxy
from overrides import overrides

from .util import PathOrStr


class FileLock(_FileLock):
    """
    This is just a subclass of the `FileLock` class from the `filelock` library, except that
    it adds an additional argument to the `__init__` method: `read_only_ok`.

    By default this flag is `False`, which an exception will be thrown when a lock
    can't be acquired due to lack of write permissions.
    But if this flag is set to `True`, a warning will be emitted instead of an error when
    the lock already exists but the lock can't be acquired because write access is blocked.
    """

    def __init__(self, lock_file: PathOrStr, timeout=-1, read_only_ok: bool = False) -> None:
        super().__init__(str(lock_file), timeout=timeout)
        self._read_only_ok = read_only_ok

    @overrides
    def acquire(self, timeout=None, poll_intervall=0.05) -> AcquireReturnProxy:
        try:
            return super().acquire(timeout=timeout, poll_intervall=poll_intervall)
        except OSError as err:
            # OSError could be a lot of different things, but what we're looking
            # for in particular are permission errors, such as:
            #  - errno 1  - EPERM  - "Operation not permitted"
            #  - errno 13 - EACCES - "Permission denied"
            #  - errno 30 - EROFS  - "Read-only file system"
            if err.errno not in (1, 13, 30):
                raise

            if os.path.isfile(self._lock_file) and self._read_only_ok:
                warnings.warn(
                    f"Lacking permissions required to obtain lock '{self._lock_file}'. "
                    "Race conditions are possible if other processes are writing to the same resource.",
                    UserWarning,
                )
            else:
                raise

    def acquire_with_updates(self, desc: str = "acquiring lock"):
        """
        Same as :meth:`acquire()`, except that when the lock cannot be immediately acquired,
        it will keep trying and print status updates as it goes.
        """
        try:
            return self.acquire(timeout=0.1)
        except Timeout:
            pass

        from .tqdm import Tqdm

        progress = Tqdm.tqdm(desc=desc, bar_format="{desc} [{elapsed}]")
        while True:
            progress.update()
            try:
                return self.acquire(timeout=1)
            except Timeout:
                continue



import copy
import inspect
from typing import Callable, Generic, TypeVar, Type, Union, Optional, Dict, Any

from .params import Params


T = TypeVar("T")


class Lazy(Generic[T]):
    """
    This class is for use when constructing objects using :class:`~tango.common.from_params.FromParams`,
    when an argument to a constructor has a `sequential dependency` with another argument to the same
    constructor.

    For example, in a ``Trainer`` class you might want to take a ``Model`` and an ``Optimizer`` as arguments,
    but the ``Optimizer`` needs to be constructed using the parameters from the ``Model``. You can give
    the type annotation ``Lazy[Optimizer]`` to the optimizer argument, then inside the constructor
    call ``optimizer.construct(parameters=model.parameters)``.

    This is only recommended for use when you have registered a ``@classmethod`` as the constructor
    for your class, instead of using ``__init__``.  Having a ``Lazy[]`` type annotation on an argument
    to an ``__init__`` method makes your class completely dependent on being constructed using the
    ``FromParams`` pipeline, which is not a good idea.

    The actual implementation here is incredibly simple; the logic that handles the lazy
    construction is actually found in ``FromParams``, where we have a special case for a ``Lazy`` type
    annotation.

    Examples
    --------

    ::

        @classmethod
        def my_constructor(
            cls,
            some_object: Lazy[MyObject],
            optional_object: Lazy[MyObject] = None,
            # or:
            #  optional_object: Optional[Lazy[MyObject]] = None,
            optional_object_with_default: Optional[Lazy[MyObject]] = Lazy(MyObjectDefault),
            required_object_with_default: Lazy[MyObject] = Lazy(MyObjectDefault),
        ) -> MyClass:
            obj1 = some_object.construct()
            obj2 = None if optional_object is None else optional_object.construct()
            obj3 = None optional_object_with_default is None else optional_object_with_default.construct()
            obj4 = required_object_with_default.construct()

    """

    def __init__(
        self,
        constructor: Union[Type[T], Callable[..., T]],
        params: Optional[Params] = None,
        constructor_extras: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        self._constructor = constructor
        self._params = params or Params({})
        self._constructor_extras = constructor_extras or {}
        self._constructor_extras.update(kwargs)

    @property
    def constructor(self) -> Callable[..., T]:
        if inspect.isclass(self._constructor):

            def constructor_to_use(**kwargs):
                return self._constructor.from_params(  # type: ignore[union-attr]
                    copy.deepcopy(self._params),
                    **kwargs,
                )

            return constructor_to_use
        else:
            return self._constructor

    def construct(self, **kwargs) -> T:
        """
        Call the constructor to create an instance of ``T``.
        """
        # If there are duplicate keys between self._constructor_extras and kwargs,
        # this will overwrite the ones in self._constructor_extras with what's in kwargs.
        contructor_kwargs = {**self._constructor_extras, **kwargs}
        return self.constructor(**contructor_kwargs)



import logging
import random
import re
from abc import abstractmethod
from pathlib import Path
from typing import (
    Optional,
    Any,
    Dict,
    Type,
    Union,
    cast,
    TypeVar,
    Generic,
    Iterator,
    Callable,
    TYPE_CHECKING,
)

try:
    from typing import get_origin, get_args  # type: ignore
except ImportError:

    def get_origin(tp):  # type: ignore
        return getattr(tp, "__origin__", None)

    def get_args(tp):  # type: ignore
        return getattr(tp, "__args__", ())


from tango.common.det_hash import det_hash
from tango.common.registrable import Registrable
from tango.common.params import Params
from tango.common.exceptions import ConfigurationError
from tango.common.from_params import (
    pop_and_construct_arg,
    infer_method_params,
    infer_constructor_params,
)
from tango.common.logging import TangoLogger
from tango.format import Format, DillFormat
from tango.step_cache import StepCache

if TYPE_CHECKING:
    from tango.executor import Executor

logger = logging.getLogger(__name__)

_version_re = re.compile("""^[a-zA-Z0-9]+$""")

T = TypeVar("T")


class Step(Registrable, Generic[T]):
    """
    This class defines one step in your experiment. To write your own step, just derive from this class
    and overwrite the :meth:`run()` method. The :meth:`run()` method must have parameters with type hints.

    ``Step.__init__()`` takes all the arguments we want to run the step with. They get passed
    to :meth:`run()` (almost) as they are. If the arguments are other instances of ``Step``, those
    will be replaced with the step's results before calling :meth:`run()`. Further, there are four special
    parameters:

    * ``step_name`` contains an optional human-readable name for the step. This name is used for
      error messages and the like, and has no consequence on the actual computation.
    * ``cache_results`` specifies whether the results of this step should be cached. If this is
      ``False``, the step is recomputed every time it is needed. If this is not set at all,
      we cache if the step is marked as :attr:`DETERMINISTIC`, and we don't cache otherwise.
    * ``step_format`` gives you a way to override the step's default format (which is given in :attr:`FORMAT`).
    * ``only_if_needed`` specifies whether we can skip this step if no other step depends on it. The
      default for this setting is to set it for all steps that don't have an explicit name.
    """

    DETERMINISTIC: bool = False
    """This describes whether this step can be relied upon to produce the same results every time
    when given the same inputs. If this is ``False``, the step can't be cached, and neither can any
    step that depends on it."""

    CACHEABLE: Optional[bool] = None
    """This provides a direct way to turn off caching. For example, a step that reads a HuggingFace
    dataset doesn't need to be cached, because HuggingFace datasets already have their own caching
    mechanism. But it's still a deterministic step, and all following steps are allowed to cache.
    If it is ``None``, the step figures out by itself whether it should be cacheable or not."""

    VERSION: Optional[str] = None
    """This is optional, but recommended. Specifying a version gives you a way to tell Tango that
    a step has changed during development, and should now be recomputed. This doesn't invalidate
    the old results, so when you revert your code, the old cache entries will stick around and be
    picked up."""

    FORMAT: Format = DillFormat("gz")
    """This specifies the format the results of this step will be serialized in. See the documentation
    for :class:`~tango.format.Format` for details."""

    def __init__(
        self,
        step_name: Optional[str] = None,
        cache_results: Optional[bool] = None,
        step_format: Optional[Format] = None,
        only_if_needed: Optional[bool] = None,
        **kwargs,
    ):
        self.logger = cast(TangoLogger, logging.getLogger(self.__class__.__name__))

        if self.VERSION is not None:
            assert _version_re.match(
                self.VERSION
            ), f"Invalid characters in version '{self.VERSION}'"
        self.kwargs = kwargs

        if step_format is None:
            self.format = self.FORMAT
            if isinstance(self.format, type):
                self.format = self.format()
        else:
            self.format = step_format

        self.unique_id_cache: Optional[str] = None
        if step_name is None:
            self.name = self.unique_id
        else:
            self.name = step_name

        if cache_results is True:
            if not self.CACHEABLE:
                raise ConfigurationError(
                    f"Step {self.name} is configured to use the cache, but it's not a cacheable step."
                )
            if not self.DETERMINISTIC:
                logger.warning(
                    f"Step {self.name} is going to be cached despite not being deterministic."
                )
            self.cache_results = True
        elif cache_results is False:
            self.cache_results = False
        elif cache_results is None:
            c = (self.DETERMINISTIC, self.CACHEABLE)
            if c == (False, None):
                self.cache_results = False
            elif c == (True, None):
                self.cache_results = True
            elif c == (False, False):
                self.cache_results = False
            elif c == (True, False):
                self.cache_results = False
            elif c == (False, True):
                logger.warning(
                    f"Step {self.name} is set to be cacheable despite not being deterministic."
                )
                self.cache_results = True
            elif c == (True, True):
                self.cache_results = True
            else:
                assert False, "Step.DETERMINISTIC or step.CACHEABLE are set to an invalid value."
        else:
            raise ConfigurationError(
                f"Step {self.name}'s cache_results parameter is set to an invalid value."
            )

        if step_name is None:
            self.only_if_needed = True
        else:
            self.only_if_needed = not self.cache_results
        if only_if_needed is not None:
            self.only_if_needed = only_if_needed

        self.work_dir_for_run: Optional[
            Path
        ] = None  # This is set only while the run() method runs.

        self._executor: Optional["Executor"] = None

    @classmethod
    def from_params(  # type: ignore[override]
        cls: Type["Step"],
        params: Union[Params, dict, str],
        constructor_to_call: Callable[..., "Step"] = None,
        constructor_to_inspect: Union[Callable[..., "Step"], Callable[["Step"], None]] = None,
        existing_steps: Optional[Dict[str, "Step"]] = None,
        step_name: Optional[str] = None,
        **extras,
    ) -> "Step":
        # Why do we need a custom from_params? Step classes have a run() method that takes all the
        # parameters necessary to perform the step. The __init__() method of the step takes those
        # same parameters, but each of them could be wrapped in another Step instead of being
        # supplied directly. from_params() doesn't know anything about these shenanigans, so
        # we have to supply the necessary logic here.

        if constructor_to_call is not None:
            raise ConfigurationError(
                f"{cls.__name__}.from_params cannot be called with a constructor_to_call."
            )
        if constructor_to_inspect is not None:
            raise ConfigurationError(
                f"{cls.__name__}.from_params cannot be called with a constructor_to_inspect."
            )

        if existing_steps is None:
            existing_steps = {}

        if isinstance(params, str):
            params = Params({"type": params})

        if not isinstance(params, Params):
            if isinstance(params, dict):
                params = Params(params)
            else:
                raise ConfigurationError(
                    "from_params was passed a ``params`` object that was not a ``Params``. This probably "
                    "indicates malformed parameters in a configuration file, where something that "
                    "should have been a dictionary was actually a list, or something else. "
                    f"This happened when constructing an object of type {cls}."
                )

        as_registrable = cast(Type[Registrable], cls)
        if "type" in params and params["type"] not in as_registrable.list_available():
            as_registrable.search_modules(params["type"])
        choice = params.pop_choice(
            "type", choices=as_registrable.list_available(), default_to_first_choice=True
        )
        subclass, constructor_name = as_registrable.resolve_class_name(choice)
        if not issubclass(subclass, Step):
            # This can happen if `choice` is a fully qualified name.
            raise ConfigurationError(
                f"Tried to make a Step of type {choice}, but ended up with a {subclass}."
            )

        parameters = infer_method_params(subclass, subclass.run, infer_kwargs=False)
        del parameters["self"]
        init_parameters = infer_constructor_params(subclass)
        del init_parameters["self"]
        del init_parameters["kwargs"]
        parameter_overlap = parameters.keys() & init_parameters.keys()
        assert len(parameter_overlap) <= 0, (
            f"If this assert fails it means that you wrote a Step with a run() method that takes one of the "
            f"reserved parameters ({', '.join(init_parameters.keys())})"
        )
        parameters.update(init_parameters)

        kwargs: Dict[str, Any] = {}
        accepts_kwargs = False
        for param_name, param in parameters.items():
            if param.kind == param.VAR_KEYWORD:
                # When a class takes **kwargs we store the fact that the method allows extra keys; if
                # we get extra parameters, instead of crashing, we'll just pass them as-is to the
                # constructor, and hope that you know what you're doing.
                accepts_kwargs = True
                continue

            explicitly_set = param_name in params
            constructed_arg = pop_and_construct_arg(
                subclass.__name__,
                param_name,
                param.annotation,
                param.default,
                params,
                **extras,
            )

            # If the param wasn't explicitly set in `params` and we just ended up constructing
            # the default value for the parameter, we can just omit it.
            # Leaving it in can cause issues with **kwargs in some corner cases, where you might end up
            # with multiple values for a single parameter (e.g., the default value gives you lazy=False
            # for a dataset reader inside **kwargs, but a particular dataset reader actually hard-codes
            # lazy=True - the superclass sees both lazy=True and lazy=False in its constructor).
            if explicitly_set or constructed_arg is not param.default:
                kwargs[param_name] = constructed_arg

        if accepts_kwargs:
            kwargs.update(params)
        else:
            params.assert_empty(subclass.__name__)

        return subclass(step_name=step_name, **kwargs)

    @abstractmethod
    def run(self, **kwargs) -> T:
        """Execute the step's action."""
        raise NotImplementedError()

    def _run_with_work_dir(self, cache: StepCache) -> T:
        if self.work_dir_for_run is not None:
            raise ValueError("You can only run a Step's run() method once at a time.")

        logger.info("Starting run for step %s of type %s", self.name, self.__class__.__name__)

        if self.DETERMINISTIC:
            random.seed(784507111)

        step_dir = cache.path_for_step(self)
        self.work_dir_for_run = step_dir / "work"
        try:
            self.work_dir_for_run.mkdir(exist_ok=True, parents=True)
            return self.run(**self.kwargs)
        finally:
            # No cleanup, as we want to keep the directory for restarts or serialization.
            self.work_dir_for_run = None

    @property
    def work_dir(self) -> Path:
        """
        Returns a work directory that a step can use while its ``run()`` method runs.

        This directory stays around across restarts. You cannot assume that it is empty when your
        step runs, but you can use it to store information that helps you restart a step if it
        got killed half-way through the last time it ran."""
        if self.work_dir_for_run is None:
            raise ValueError("You can only call this method while the step is running.")
        return self.work_dir_for_run

    @property
    def executor(self) -> "Executor":
        """
        The executor being used to run this step.

        This should only be called during the step's ``run()`` method.
        """
        if self._executor is None:
            raise ValueError("Step.executor cannot be called outside of step execution!")
        else:
            return self._executor

    def result(self, cache: StepCache) -> T:
        """Returns the result of this step. If the results are cached, it returns those. Otherwise it
        runs the step and returns the result from there."""
        if self in cache:
            return cache[self]

        result = self._run_with_work_dir(cache)
        if self.cache_results:
            cache[self] = result
            if hasattr(result, "__next__"):
                assert isinstance(result, Iterator)
                # Caching the iterator will consume it, so we write it to the cache and then read from the cache
                # for the return value.
                return cache[self]
        return result

    def det_hash_object(self) -> Any:
        return self.unique_id

    @property
    def unique_id(self) -> str:
        """Returns the unique ID for this step.

        Unique IDs are of the shape ``$class_name-$version-$hash``, where the hash is the hash of the
        inputs for deterministic steps, and a random string of characters for non-deterministic ones."""
        if self.unique_id_cache is None:
            self.unique_id_cache = self.__class__.__name__
            if self.VERSION is not None:
                self.unique_id_cache += "-"
                self.unique_id_cache += self.VERSION

            self.unique_id_cache += "-"
            if self.DETERMINISTIC:
                self.unique_id_cache += det_hash(
                    (
                        (self.format.__class__.__module__, self.format.__class__.__qualname__),
                        self.format.VERSION,
                        self.kwargs,
                    )
                )[:32]
            else:
                self.unique_id_cache += det_hash(random.getrandbits((58 ** 32).bit_length()))[:32]

        return self.unique_id_cache

    def __hash__(self):
        return hash(self.unique_id)

    def __eq__(self, other):
        if isinstance(other, Step):
            return self.unique_id == other.unique_id
        else:
            return False



"""
Components for Tango integration with `PyTorch <https://pytorch.org/>`_.

These include a basic training loop :class:`~tango.step.Step` and registrable versions
of many ``torch`` classes, such :class:`torch.optim.Optimizer` and :class:`torch.utils.data.DataLoader`.

Example: training a model
-------------------------

Let's look a simple example of training a model.

We'll make a very basic regression model and generate some fake data to train on.
First, the setup:

.. testcode::

    import torch
    import torch.nn as nn

    from tango.common.dataset_dict import DatasetDict
    from tango.step import Step
    from tango.integrations.torch import Model

Now let's build and register our model:

.. testcode::

    @Model.register("basic_regression")
    class BasicRegression(Model):
        def __init__(self):
            super().__init__()
            self.linear = nn.Linear(10, 1)
            self.sigmoid = nn.Sigmoid()
            self.mse = nn.MSELoss()

        def forward(self, x, y=None):
            pred = self.sigmoid(self.linear(x))
            out = {"pred": pred}
            if y is not None:
                out["loss"] = self.mse(pred, y)
            return out

        def _to_params(self):
            return {}

Lastly, we'll need a step to generate data:

.. testcode::

    @Step.register("generate_data")
    class GenerateData(Step):
        DETERMINISTIC = True
        CACHEABLE = False

        def run(self) -> DatasetDict:
            torch.manual_seed(1)
            return DatasetDict(
                {
                    "train": [{"x": torch.rand(10), "y": torch.rand(1)} for _ in range(64)],
                    "validation": [{"x": torch.rand(10), "y": torch.rand(1)} for _ in range(32)],
                }
            )

You could then run this experiment with a config that looks like this:

.. testcode::

    config = {
        "steps": {
            "data": {
                "type": "generate_data",
            },
            "train": {
                "type": "torch::train",
                "model": {
                    "type": "basic_regression",
                },
                "dataset_dict": {
                    "type": "ref",
                    "ref": "data",
                },
                "train_dataloader": {
                    "batch_size": 8,
                    "shuffle": True,
                },
                "optimizer": {
                    "type": "Adam",
                },
                "validation_split": "validation",
                "validation_dataloader": {
                    "batch_size": 8,
                    "shuffle": False,
                },
                "train_steps": 100,
                "validate_every": 10,
                "checkpoint_every": 10,
                "log_every": 1,
            }
        }
    }

.. testcode::
    :hide:

    import os
    from tango.common.testing import run_experiment
    from tango.common.registrable import Registrable

    # Don't cache results, otherwise we'll have a pickling error.
    config["steps"]["train"]["cache_results"] = False
    with run_experiment(config) as run_dir:
        assert (run_dir / "step_cache").is_dir()
    # Restore state of registry.
    del Registrable._registry[Step]["generate_data"]
    del Registrable._registry[Model]["basic_regression"]

For example,

.. code-block::

    tango run config.jsonnet -i my_package -d /tmp/train

would produce the following output:

.. testoutput::

    ● Starting run for "data"
    ✓ Finished run for "data"
    ● Starting run for "train"
    ✓ Finished run for "train"

"""

__all__ = [
    "TorchFormat",
    "TorchTrainStep",
    "Optimizer",
    "LRScheduler",
    "Model",
    "DataLoader",
    "DataCollator",
    "Sampler",
    "ConcatTensorDictsCollator",
]

from .data import DataLoader, Sampler, DataCollator, ConcatTensorDictsCollator
from .format import TorchFormat
from .model import Model
from .optim import Optimizer, LRScheduler
from .train import TorchTrainStep



import collections
from dataclasses import dataclass
import logging
import weakref
from abc import abstractmethod
from pathlib import Path
from typing import (
    Optional,
    Any,
    TypeVar,
    MutableMapping,
    OrderedDict,
    TYPE_CHECKING,
)

try:
    from typing import get_origin, get_args  # type: ignore
except ImportError:

    def get_origin(tp):  # type: ignore
        return getattr(tp, "__origin__", None)

    def get_args(tp):  # type: ignore
        return getattr(tp, "__args__", ())


from tango.common.from_params import FromParams
from tango.common.registrable import Registrable
from tango.common.util import PathOrStr

if TYPE_CHECKING:
    from tango.step import Step

logger = logging.getLogger(__name__)


T = TypeVar("T")


class StepCache(Registrable):
    """
    This is a mapping from instances of :class:`~tango.step.Step` to the results of that step.
    """

    default_implementation = "local"
    """
    The default implementation is :class:`LocalStepCache`.
    """

    def __init__(self, dir: PathOrStr):
        self.dir = Path(dir)
        self.dir.mkdir(parents=True, exist_ok=True)

    def __contains__(self, step: object) -> bool:
        """This is a generic implementation of ``__contains__``. If you are writing your own
        ``StepCache``, you might want to write a faster one yourself."""
        from tango.step import Step

        if not isinstance(step, Step):
            return False
        try:
            self.__getitem__(step)
            return True
        except KeyError:
            return False

    @abstractmethod
    def __getitem__(self, step: "Step") -> Any:
        """Returns the results for the given step."""
        raise NotImplementedError()

    @abstractmethod
    def __setitem__(self, step: "Step", value: Any) -> None:
        """Writes the results for the given step. Throws an exception if the step is already cached."""
        raise NotImplementedError()

    @abstractmethod
    def __len__(self) -> int:
        """Returns the number of results saved in this cache."""
        raise NotImplementedError()

    def path_for_step(self, step: "Step") -> Path:
        """Steps that can be restarted (like a training job that gets interrupted half-way through)
        must save their state somewhere. A :class:`StepCache` can help by providing a suitable location
        in this method."""
        return self.dir / step.unique_id


@StepCache.register("local")
class LocalStepCache(StepCache):
    """
    This is a :class:`StepCache` that stores its results on disk, in the location given in ``dir``.

    Every cached step gets a directory under ``dir`` with that step's :attr:`~tango.step.Step.unique_id`.
    In that directory we store the results themselves in some format according to the step's
    :attr:`~tango.step.Step.FORMAT`,
    and we also write a ``cache-metadata.json`` file that stores some metadata. The presence of
    ``cache-metadata.json`` signifies that the cache entry is complete and has been written successfully.

    .. tip::
        Registered as :class:`StepCache` under the name "local".

    """

    LRU_CACHE_MAX_SIZE = 8

    def __init__(self, dir: PathOrStr):
        super().__init__(dir)

        # We keep an in-memory cache as well so we don't have to de-serialize stuff
        # we happen to have in memory already.
        self.weak_cache: MutableMapping[str, Any] = weakref.WeakValueDictionary()

        # Not all Python objects can be referenced weakly, and even if they can they
        # might get removed too quickly, so we also keep an LRU cache.
        self.strong_cache: OrderedDict[str, Any] = collections.OrderedDict()

    def _add_to_cache(self, key: str, o: Any) -> None:
        if hasattr(o, "__next__"):
            # We never cache iterators, because they are mutable, storing their current position.
            return

        self.strong_cache[key] = o
        self.strong_cache.move_to_end(key)
        while len(self.strong_cache) > self.LRU_CACHE_MAX_SIZE:
            del self.strong_cache[next(iter(self.strong_cache))]

        try:
            self.weak_cache[key] = o
        except TypeError:
            pass  # Many native Python objects cannot be referenced weakly, and they throw TypeError when you try

    def _get_from_cache(self, key: str) -> Optional[Any]:
        result = self.strong_cache.get(key)
        if result is not None:
            self.strong_cache.move_to_end(key)
            return result
        try:
            return self.weak_cache[key]
        except KeyError:
            return None

    def __contains__(self, step: object) -> bool:
        from tango.step import Step

        if isinstance(step, Step):
            key = step.unique_id
            if key in self.strong_cache:
                return True
            if key in self.weak_cache:
                return True
            metadata_file = self.path_for_step(step) / "cache-metadata.json"
            return metadata_file.exists()
        else:
            return False

    def __getitem__(self, step: "Step") -> Any:
        key = step.unique_id
        result = self._get_from_cache(key)
        if result is None:
            if step not in self:
                raise KeyError(step)
            result = step.format.read(self.path_for_step(step))
            self._add_to_cache(key, result)
        return result

    def __setitem__(self, step: "Step", value: Any) -> None:
        location = self.path_for_step(step)
        location.mkdir(parents=True, exist_ok=True)

        metadata_location = location / "cache-metadata.json"
        if metadata_location.exists():
            raise ValueError(f"{metadata_location} already exists! Will not overwrite.")
        temp_metadata_location = metadata_location.with_suffix(".temp")

        try:
            step.format.write(value, location)
            metadata = CacheMetadata(step=step.unique_id, checksum=step.format.checksum(location))
            metadata.to_params().to_file(temp_metadata_location)
            self._add_to_cache(step.unique_id, value)
            temp_metadata_location.rename(metadata_location)
        except:  # noqa: E722
            try:
                temp_metadata_location.unlink()
            except FileNotFoundError:
                pass
            raise

    def __len__(self) -> int:
        return sum(1 for _ in self.dir.glob("*/cache-metadata.json"))


@dataclass
class CacheMetadata(FromParams):
    step: str
    checksum: str



"""
A Python library for choreographing your machine learning research.
"""

__all__ = [
    "Format",
    "DillFormat",
    "DillFormatIterator",
    "Executor",
    "JsonFormat",
    "JsonFormatIterator",
    "Step",
    "StepCache",
    "LocalStepCache",
]

from tango.executor import Executor, SimpleExecutor
from tango.format import Format, DillFormat, DillFormatIterator, JsonFormat, JsonFormatIterator
from tango.step import Step
from tango.step_cache import StepCache, LocalStepCache
from tango.step_graph import StepGraph, StepStub



"""
Components for Tango integration with `🤗 Datasets <https://huggingface.co/docs/datasets/>`_.
"""


from typing import Union

import datasets
from datasets import DatasetDict, Dataset, IterableDatasetDict, IterableDataset

from tango.step import Step


__all__ = ["LoadDataset"]


@Step.register("datasets::load")
class LoadDataset(Step):
    """
    This step loads a `HuggingFace dataset <https://huggingface.co/datasets>`_.

    .. tip::

        Registered as a :class:`~tango.step.Step` under the name "datasets::load".

    Examples
    --------

    .. testsetup::

        from tango import Step

    .. testcode::

        load_step = Step.from_params({
            "type": "datasets::load",
            "path": "lhoestq/test",
        })

    """

    DETERMINISTIC = True
    VERSION = "001"
    CACHEABLE = False  # These are already cached by huggingface.

    def run(self, path: str, **kwargs) -> Union[DatasetDict, Dataset, IterableDatasetDict, IterableDataset]:  # type: ignore
        """
        Loads a HuggingFace dataset.

        ``path`` is the canonical name or path to the dataset. Additional key word arguments
        are passed as-is to :func:`datasets.load_dataset()`.
        """
        return datasets.load_dataset(path, **kwargs)



import os
import logging

from .util import _parse_bool


class TangoLogger(logging.Logger):
    """
    A custom subclass of 'logging.Logger' that keeps a set of messages to
    implement {debug,info,etc.}_once() methods.
    """

    def __init__(self, name):
        super().__init__(name)
        self._seen_msgs = set()

    def debug_once(self, msg, *args, **kwargs):
        if msg not in self._seen_msgs:
            self.debug(msg, *args, **kwargs)
            self._seen_msgs.add(msg)

    def info_once(self, msg, *args, **kwargs):
        if msg not in self._seen_msgs:
            self.info(msg, *args, **kwargs)
            self._seen_msgs.add(msg)

    def warning_once(self, msg, *args, **kwargs):
        if msg not in self._seen_msgs:
            self.warning(msg, *args, **kwargs)
            self._seen_msgs.add(msg)

    def error_once(self, msg, *args, **kwargs):
        if msg not in self._seen_msgs:
            self.error(msg, *args, **kwargs)
            self._seen_msgs.add(msg)

    def critical_once(self, msg, *args, **kwargs):
        if msg not in self._seen_msgs:
            self.critical(msg, *args, **kwargs)
            self._seen_msgs.add(msg)


logging.setLoggerClass(TangoLogger)
logger = logging.getLogger(__name__)


FILE_FRIENDLY_LOGGING: bool = _parse_bool(os.environ.get("FILE_FRIENDLY_LOGGING", False))
"""
If this flag is set to ``True``, we add newlines to tqdm output, even on an interactive terminal, and we slow
down tqdm's output to only once every 10 seconds.

By default, it is set to ``False``.
"""



from itertools import islice
import os
from pathlib import Path
import random
import tempfile
import typing as t

from more_itertools import chunked
import numpy as np
import torch
from torch import Tensor
import torch.distributed as dist
import torch.nn as nn
from torch.utils.data import DistributedSampler

from .data import DataLoader
from .format import TorchFormat
from .model import Model
from .optim import Optimizer, LRScheduler
from tango.common.dataset_dict import DatasetDict
from tango.common.exceptions import ConfigurationError
from tango.common.lazy import Lazy
from tango.common.tqdm import Tqdm
from tango.format import Format
from tango.step import Step


@Step.register("torch::train")
class TorchTrainStep(Step):
    """
    A basic training loop step.

    .. tip::

        Registered as a :class:`~tango.step.Step` under the name "torch::train".
    """

    DETERMINISTIC: bool = True
    CACHEABLE = True
    FORMAT: Format = TorchFormat()

    def run(  # type: ignore[override]
        self,
        model: Lazy[Model],
        dataset_dict: DatasetDict,
        train_dataloader: Lazy[DataLoader],
        optimizer: Lazy[Optimizer],
        *,
        train_split: str = "train",
        validation_split: t.Optional[str] = None,
        lr_scheduler: t.Optional[Lazy[LRScheduler]] = None,
        validation_dataloader: t.Optional[Lazy[DataLoader]] = None,
        seed: int = 42,
        train_steps: t.Optional[int] = None,
        validation_steps: t.Optional[int] = None,
        grad_accum: int = 1,
        log_every: int = 10,
        checkpoint_every: int = 100,
        validate_every: int = 100,
        amp: bool = False,
        max_grad_norm: t.Optional[float] = None,
        devices: t.Optional[t.List[int]] = None,
        distributed_port: str = "54761",
    ) -> Model:
        """
        Run a basic training loop to train the ``model``.

        Parameters
        ----------

        model : :class:`Model`
            The model to train. It should return a ``dict`` that includes the ``loss``
            during training and validation.
        dataset_dict : :class:`~tango.common.dataset_dict.DatasetDict`
            The train and optional validation data.
        train_dataloader : :class:`DataLoader`
            The data loader that generates training batches. The batches should be :class:`dict`
            objects.
        optimizer : :class:`Optimizer`
            A PyTorch :class:`~torch.optim.Optimizer`.
        train_split : :class:`str`, optional
            The name of the data split used for training in the ``dataset_dict``.
            Default is "train".
        validation_split : :class:`str`, optional
            Optional name of the validation split in the ``dataset_dict``. Default is ``None``,
            which means no validation.
        lr_scheduler : :class:`LRScheduler`, optional
            An optional
            `learning rate scheduler <https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate>`_
            to adjust the learning rate throughout training.
        validation_dataloader : :class:`DataLoader`, optional
            An optional data loader for generating validation batches. The batches should be
            :class:`dict` objects. If not specified, but ``validation_split`` is given,
            the validation ``DataLoader`` will be constructed from the same parameters
            as the train ``DataLoader``.
        seed : :class:`int`, optional
            Used to set the RNG states at the beginning of training.
        train_steps : :class:`int`, optional
            The number of steps to train for. If not specified training will
            stop after a complete iteration through the `train_dataloader`.
        validation_steps : :class:`int`, optional
            The number of steps to validate for. If not specified validation
            will stop after a complete iteration through the `validation_dataloader`.
        grad_accum : :class:`int`, optional
            The number of gradient accumulation steps. Defaults to 1.

            .. note::
                This parameter - in conjuction with the settings of your data loader
                and the number distributed workers -
                determines the *effective batch size* of your training run.

        log_every : :class:`int`, optional
            Log every this many steps.
        checkpoint_every : :class:`int`, optional
            Save a checkpoint every this many steps.
        validate_every : :class:`int`, optional
            Run the validation loop every this many steps.
        amp : :class:`bool`, optional
            Use automatic mixed precision. Default is ``False``.
        max_grad_norm : :class:`float`, optional
            If set, gradients will be clipped to have this max norm. Default is ``None``.
        devices : ``List[int]``, optional
            The IDs of the CUDA devices to train on.
        distributed_port : :class:`str`
            The port of the distributed process group. Default = "54761".

        Returns
        -------
        :class:`Model`
            The trained model.

        """
        # Validate device(s).
        if torch.cuda.is_available():
            if devices is None:
                print("CUDA is available")
            elif all((x >= 0 for x in devices)):
                if torch.cuda.device_count() < set(devices):
                    raise ConfigurationError(
                        f"Only found {torch.cuda.device_count()} CUDA devices, "
                        f"but you specified {len(set(devices))} device IDs"
                    )
            elif not all((x == -1 for x in devices)):
                raise ConfigurationError("Invalid value for 'devices'")
        else:
            if devices and not all((x == -1 for x in devices)):
                raise ConfigurationError(
                    "CUDA not found, so only '-1' allowed for device IDs, found {devices}"
                )
            if amp:
                raise ConfigurationError("AMP requires CUDA")

        is_distributed = False
        num_workers = 1
        if devices and len(devices) > 1:
            is_distributed = True
            num_workers = len(devices)

        if is_distributed:
            import torch.multiprocessing as mp

            mp.spawn(
                _train,
                args=(
                    self.work_dir,
                    model,
                    dataset_dict,
                    train_dataloader,
                    optimizer,
                    train_split,
                    validation_split,
                    lr_scheduler,
                    validation_dataloader,
                    seed,
                    train_steps,
                    validation_steps,
                    grad_accum,
                    log_every,
                    checkpoint_every,
                    validate_every,
                    amp,
                    max_grad_norm,
                    is_distributed,
                    devices,
                    distributed_port,
                    self.executor.include_package,
                ),
                nprocs=num_workers,
            )

            self.logger.info("Loading final weights")
            final_model: Model = model.construct()
            state = torch.load(self.work_dir / "state_worker0_best.pt", map_location="cpu")
            final_model.load_state_dict(state["model"])
            return final_model
        else:
            final_model = _train(  # type: ignore[assignment]
                0,
                self.work_dir,
                model,
                dataset_dict,
                train_dataloader,
                optimizer,
                train_split=train_split,
                validation_split=validation_split,
                lr_scheduler=lr_scheduler,
                validation_dataloader=validation_dataloader,
                seed=seed,
                train_steps=train_steps,
                validation_steps=validation_steps,
                grad_accum=grad_accum,
                log_every=log_every,
                checkpoint_every=checkpoint_every,
                validate_every=validate_every,
                amp=amp,
                max_grad_norm=max_grad_norm,
                is_distributed=is_distributed,
            )
            assert final_model is not None
            return final_model


def _train(
    worker_id: int,
    work_dir: Path,
    model: Lazy[Model],
    dataset_dict: DatasetDict,
    train_dataloader: Lazy[DataLoader],
    optimizer: Lazy[Optimizer],
    train_split: str = "train",
    validation_split: t.Optional[str] = None,
    lr_scheduler: t.Optional[Lazy[LRScheduler]] = None,
    validation_dataloader: t.Optional[Lazy[DataLoader]] = None,
    seed: int = 42,
    train_steps: t.Optional[int] = None,
    validation_steps: t.Optional[int] = None,
    grad_accum: int = 1,
    log_every: int = 10,
    checkpoint_every: int = 100,
    validate_every: int = 100,
    amp: bool = False,
    max_grad_norm: t.Optional[float] = None,
    is_distributed: bool = False,
    devices: t.Optional[t.List[int]] = None,
    distributed_port: str = "54761",
    include_package: t.Optional[t.List[str]] = None,
) -> t.Optional[Model]:
    if include_package:
        from tango.common.util import import_module_and_submodules

        for package_name in include_package:
            import_module_and_submodules(package_name)

    is_local_main_process = worker_id == 0
    world_size = len(devices) if devices else 1

    # Resolve and set device.
    device: torch.device = torch.device("cpu")
    if devices:
        device_id = devices[worker_id]
        if device_id >= 0:
            device = torch.device(f"cuda:{device_id}")
            torch.cuda.set_device(device)
    elif torch.cuda.is_available():
        device = torch.device("cuda")

    # Init distributed process group.
    if is_distributed:
        assert devices
        backend = "gloo" if device == torch.device("cpu") else "nccl"
        dist.init_process_group(
            backend=backend,
            init_method=f"tcp://127.0.0.1:{distributed_port}",
            world_size=world_size,
            rank=worker_id,
        )

    grad_scaler: t.Optional[torch.cuda.amp.GradScaler] = None
    if amp:
        grad_scaler = torch.cuda.amp.GradScaler()

    state_path = work_dir / f"state_worker{worker_id}.pt"
    best_state_path = work_dir / f"state_worker{worker_id}_best.pt"

    # Construct data loaders.
    if validation_split is not None:
        if validation_dataloader is not None:
            validation_dataloader = validation_dataloader.construct(
                dataset=dataset_dict[validation_split]
            )
        else:
            validation_dataloader = train_dataloader.construct(
                dataset=dataset_dict[validation_split]
            )
    validation_dataloader: t.Optional[DataLoader] = t.cast(
        t.Optional[DataLoader], validation_dataloader
    )
    try:
        train_dataset = dataset_dict[train_split]
    except KeyError:
        raise KeyError(f"'{train_split}', available keys are {list(dataset_dict.keys())}")
    train_dataloader: DataLoader = train_dataloader.construct(dataset=train_dataset)

    if train_steps is None:
        train_steps = len(train_dataloader)
    train_steps: int = t.cast(int, train_steps)  # type: ignore[no-redef]
    if validation_dataloader is not None:
        if validation_steps is None:
            validation_steps = len(validation_dataloader)

    # Make sure we're using a DistributedSampler during distributed training.
    if is_distributed:
        if not isinstance(train_dataloader.sampler, DistributedSampler):
            raise ConfigurationError(
                "DistributedSampler is required for dataloader during distributed training, "
                f"found {type(train_dataloader.sampler)} instead."
            )
        if validation_dataloader is not None and not isinstance(
            validation_dataloader.sampler, DistributedSampler
        ):
            raise ConfigurationError(
                "DistributedSampler is required for dataloader during distributed training, "
                f"found {type(validation_dataloader.sampler)} instead."
            )

    # Check working directory to see if we should recover from a previous run.
    initial_state: t.Optional[t.Dict[str, t.Any]] = None
    if state_path.is_file():
        print(f"Recovering from previous run at {str(state_path)}")
        initial_state = torch.load(state_path)

    # Prepare model.
    model: Model = model.construct()
    if initial_state is not None:
        model.load_state_dict(initial_state["model"])
    if is_distributed:
        model = nn.parallel.DistributedDataParallel(model)
    model = model.to(device)

    # Prepare optimizer and lr scheduler.
    optimizer: Optimizer = optimizer.construct(params=model.parameters())
    if initial_state is not None:
        optimizer.load_state_dict(initial_state["optimizer"])
    if lr_scheduler is not None:
        lr_scheduler = lr_scheduler.construct(optimizer=optimizer)
        if initial_state is not None:
            lr_scheduler.load_state_dict(initial_state["scheduler"])
    lr_scheduler: t.Optional[LRScheduler] = t.cast(t.Optional[LRScheduler], lr_scheduler)

    # Set random seeds.
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)

    val_loss: t.Optional[float] = None
    best_val_loss: t.Optional[float] = None
    if initial_state is not None:
        val_loss = initial_state["val_loss"]
        best_val_loss = initial_state["best_val_loss"]

    model.train()
    training_batches = enumerate(
        islice(
            chunked(cycle_through_epochs(train_dataloader, is_distributed), grad_accum),
            train_steps,
        )
    )

    # Catch data loader up to where we left off before.
    if initial_state is not None:
        training_steps = initial_state["training_steps"]
        print(f"Catching data loader up to step {training_steps}")
        for step, batch in training_batches:
            del batch
            if step >= training_steps - 1:
                break

    if is_distributed:
        dist.barrier()

    with Tqdm.tqdm(
        training_batches,
        desc="Training",
        initial=0 if initial_state is None else initial_state["training_steps"],
        total=train_steps,
        disable=not is_local_main_process,
    ) as train_batch_iterator:
        for step, batch in train_batch_iterator:

            def save_state():
                state_path_for_step = work_dir / f"state_worker{worker_id}_step{step + 1}.pt"
                temp_state_file = tempfile.NamedTemporaryFile(
                    "w+b", dir=work_dir, delete=False, suffix=".pt"
                )
                try:
                    with Tqdm.wrapattr(
                        temp_state_file,
                        "write",
                        desc="Saving checkpoint",
                        leave=False,
                        disable=not is_local_main_process,
                    ) as f:
                        torch.save(
                            {
                                "optimizer": optimizer.state_dict(),  # type: ignore[attr-defined]
                                "scheduler": None
                                if lr_scheduler is None
                                else lr_scheduler.state_dict(),  # type: ignore[attr-defined]
                                "model": model.module.state_dict()  # type: ignore[attr-defined]
                                if is_distributed
                                else model.state_dict(),  # type: ignore[attr-defined]
                                "training_steps": step + 1,
                                "val_loss": val_loss,
                                "best_val_loss": best_val_loss,
                            },
                            f,
                        )
                    temp_state_file.close()
                    os.replace(temp_state_file.name, state_path_for_step)

                    # Link to most recent state path.
                    if state_path.is_symlink():
                        state_path.unlink()
                    state_path.symlink_to(state_path_for_step)

                    # Link to best state path.
                    if (
                        val_loss is not None
                        and best_val_loss is not None
                        and val_loss <= best_val_loss
                    ):
                        if best_state_path.is_symlink():
                            best_state_path.unlink()
                        best_state_path.symlink_to(state_path_for_step)
                finally:
                    if os.path.exists(temp_state_file.name):
                        os.remove(temp_state_file.name)

            optimizer.zero_grad()
            batch_loss = 0.0
            for micro_batch in batch:
                # Move tensors to right device.
                micro_batch = move_to_device(micro_batch, device)

                # Get loss.
                with torch.cuda.amp.autocast(enabled=amp):
                    outputs = model(**micro_batch)
                    micro_batch_loss = outputs["loss"] / len(batch)

                if torch.isnan(micro_batch_loss):
                    raise ValueError("nan loss encountered")
                batch_loss += micro_batch_loss.detach().item()

                # Calculate gradients.
                if grad_scaler is not None:
                    grad_scaler.scale(micro_batch_loss).backward()
                else:
                    micro_batch_loss.backward()

                # Clean up in case it saves memory.
                del micro_batch
                del outputs
                del micro_batch_loss

            del batch

            # Unscale gradients.
            if grad_scaler is not None:
                grad_scaler.unscale_(optimizer)

            # Clip gradients.
            if max_grad_norm is not None:
                torch.nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)

            # Take optimizer step.
            if grad_scaler is not None:
                grad_scaler.step(optimizer)
                grad_scaler.update()
            else:
                optimizer.step()

            # Adjust LR schedule.
            if lr_scheduler is not None:
                lr_scheduler.step()

            should_log_this_step = step % log_every == 0 or step == train_steps - 1
            should_checkpoint_this_step = (
                step > 0 and step % checkpoint_every == 0
            ) or step == train_steps - 1
            should_validate_this_step = validation_dataloader is not None and (
                (step > 0 and step % validate_every == 0) or step == train_steps - 1
            )

            # Gather average loss across all workers.
            if (should_log_this_step or should_validate_this_step) and is_distributed:
                batch_loss_tensor = torch.tensor(batch_loss, device=device)
                dist.all_reduce(batch_loss_tensor)
                batch_loss = batch_loss_tensor.item() / world_size

            # Update progress bar.
            if is_local_main_process and should_log_this_step:
                if val_loss is not None:
                    train_batch_iterator.set_postfix(
                        batch_loss=batch_loss,
                        val_loss=val_loss,
                    )
                else:
                    train_batch_iterator.set_postfix(batch_loss=batch_loss)

            # Validate.
            if should_validate_this_step:
                assert validation_dataloader is not None
                assert validation_steps is not None

                # Prepare model for validation.
                model.eval()
                optimizer.zero_grad()  # Not strictly necessary.

                running_loss = 0.0
                with Tqdm.tqdm(
                    islice(validation_dataloader, validation_steps),
                    desc="Validating",
                    total=validation_steps,
                    leave=False,
                    disable=not is_local_main_process,
                ) as val_batch_iterator:
                    for val_step, val_batch in enumerate(val_batch_iterator):
                        # Move tensors to right device.
                        val_batch = move_to_device(val_batch, device)

                        # Get loss.
                        with torch.cuda.amp.autocast(enabled=amp):
                            with torch.inference_mode():
                                outputs = model(**val_batch)
                        loss = outputs["loss"]

                        running_loss += loss.item()
                        val_loss = running_loss / (val_step + 1)

                        should_log_this_step = (
                            val_step % log_every == 0 or val_step == validation_steps - 1
                        )

                        # Average loss across all workers.
                        if is_distributed and should_log_this_step:
                            val_loss_tensor = torch.tensor(val_loss, device=device)
                            dist.all_reduce(val_loss_tensor)
                            val_loss = val_loss_tensor.item() / world_size

                        # Update progress bar.
                        if is_local_main_process and should_log_this_step:
                            val_batch_iterator.set_postfix(loss=val_loss)

                        # Clean up.
                        del val_batch
                        del outputs
                        del loss

                # Reset model to train mode.
                model.train()

                # Update progress bar again.
                if is_local_main_process:
                    train_batch_iterator.set_postfix(
                        batch_loss=batch_loss,
                        val_loss=val_loss,
                    )

                assert val_loss is not None
                if best_val_loss is None:
                    best_val_loss = val_loss
                elif val_loss <= best_val_loss:
                    best_val_loss = val_loss

            # Checkpoint.
            if should_checkpoint_this_step:
                save_state()

    if is_distributed:
        dist.barrier()

    if not is_distributed:
        return model
    else:
        return None


def cycle_through_epochs(dataloader: DataLoader, is_distributed: bool):
    epoch = 0
    while True:
        if is_distributed and isinstance(dataloader.sampler, DistributedSampler):
            dataloader.sampler.set_epoch(epoch)
        for batch in dataloader:
            yield batch
        epoch += 1


T = t.TypeVar("T")


def move_to_device(o: T, device: torch.device) -> T:
    if isinstance(o, Tensor):
        return o.to(device)
    elif isinstance(o, dict):
        return {k: move_to_device(v, device) for k, v in o.items()}  # type: ignore[return-value]
    elif isinstance(o, list):
        return [move_to_device(x, device) for x in o]  # type: ignore[return-value]
    elif isinstance(o, tuple):
        return tuple((move_to_device(x, device) for x in o))  # type: ignore[return-value]
    else:
        return o



from collections import OrderedDict
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Dict, Any, Set, Union, List

from tango.common.exceptions import ConfigurationError
from tango.common.params import Params


@dataclass
class StepStub:
    """
    Stub for a :class:`~tango.step.Step`.
    """

    name: str
    """
    The name of the step.
    """

    config: Dict[str, Any]
    """
    The configuration for the step.
    """

    dependencies: Set[str]
    """
    The other steps that this step directly depends on.
    """


class StepGraph(Sequence):
    """
    Represents an experiment as a graph.

    It can be treated as either a :class:`~collections.abc.Mapping` of step names (``str``)
    to :class:`StepStub`, or simply a :class:`~collections.abc.Sequence` of :class:`StepStub`.

    When treated as a sequence, it can be assumed that no step in the sequence depends on a step
    before it.
    """

    def __init__(self, steps: Union[Dict[str, Any], Params]) -> None:
        if isinstance(steps, Params):
            steps = steps.as_dict()
        remaining_steps_to_sort: Dict[str, StepStub] = {}
        for step_name, step_config in steps.items():
            dependencies = self._parse_direct_step_dependencies(step_config)
            remaining_steps_to_sort[step_name] = StepStub(
                name=step_name, config=step_config, dependencies=dependencies
            )

        # These are the steps in the order that they should run.
        self._ordered_steps: Dict[str, StepStub] = OrderedDict()
        for _ in range(len(remaining_steps_to_sort)):
            # Go through sorted to ensure this is deterministic.
            for step_name in sorted(remaining_steps_to_sort.keys()):
                step_stub = remaining_steps_to_sort[step_name]
                for ref in step_stub.dependencies:
                    if ref not in self._ordered_steps:
                        # Step depends an other later step, so it needs to wait.
                        break
                else:
                    self._ordered_steps[step_name] = step_stub
                    remaining_steps_to_sort.pop(step_name)
                    break

        # Validate the graph.
        if remaining_steps_to_sort:
            err_msgs: List[str] = []
            for step_name, step_stub in remaining_steps_to_sort.items():
                for ref in step_stub.dependencies:
                    if ref not in self._ordered_steps:
                        err_msgs.append(f"Can't resolve dependency {ref} for {step_name}")

            raise ConfigurationError("Invalid step graph:\n- " + "\n- ".join(err_msgs))

    def __getitem__(self, key: Union[str, int]) -> StepStub:  # type: ignore[override]
        """
        Get the stub corresponding to ``key``.
        """
        if isinstance(key, str):
            return self._ordered_steps[key]
        else:
            return self._ordered_steps[list(self._ordered_steps.keys())[key]]

    def __len__(self) -> int:
        """
        The number of steps in the experiment.
        """
        return len(self._ordered_steps)

    @staticmethod
    def _parse_direct_step_dependencies(o: Any) -> Set[str]:
        dependencies: Set[str] = set()
        if isinstance(o, (list, tuple, set)):
            for item in o:
                dependencies = dependencies | StepGraph._parse_direct_step_dependencies(item)
        elif isinstance(o, dict):
            if set(o.keys()) == {"type", "ref"}:
                dependencies.add(o["ref"])
            else:
                for value in o.values():
                    dependencies = dependencies | StepGraph._parse_direct_step_dependencies(value)
        elif not isinstance(o, (bool, str, int, float)):
            raise ValueError(o)
        return dependencies



import logging
import os
from typing import Optional, Union, List

import click
from click_help_colors import HelpColorsCommand, HelpColorsGroup

from tango.version import VERSION


@click.group(
    cls=HelpColorsGroup,
    help_options_color="green",
    help_headers_color="yellow",
    context_settings={"max_content_width": 115},
)
@click.version_option(version=VERSION)
@click.option(
    "--log-level",
    help="Set the global log level.",
    type=click.Choice(["debug", "info", "warning", "error"], case_sensitive=False),
    show_choices=True,
    default="info",
)
@click.option(
    "--no-logging",
    is_flag=True,
    help="Disable logging altogether.",
)
def main(log_level, no_logging):
    if not no_logging:
        level = logging._nameToLevel[log_level.upper()]
        logging.basicConfig(
            format="[%(asctime)s %(levelname)s %(name)s] %(message)s",
            level=level,
        )
        # filelock emits too many messages, so tell it to be quiet unless it has something
        # important to say.
        logging.getLogger("filelock").setLevel(max(level, logging.WARNING))

    from tango.common.util import install_sigterm_handler

    # We want to be able to catch SIGTERM signals in addition to SIGINT (keyboard interrupt).
    install_sigterm_handler()


@main.command(
    cls=HelpColorsCommand,
    help_options_color="green",
    help_headers_color="yellow",
    context_settings={"max_content_width": 115},
)
@click.argument(
    "experiment",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@click.option(
    "-d",
    "--directory",
    type=click.Path(file_okay=False),
    help="""The directory in which to save the results of each step. If not specified,
    a named temporary directory will be created.""",
    default=None,
)
@click.option(
    "-o",
    "--overrides",
    type=str,
    help="""A JSON(NET) string used to override fields in the experiment config.
    Use dot syntax to specify nested fields.""",
)
@click.option(
    "-i",
    "--include-package",
    type=str,
    help="Python packages or modules to import for tango components.",
    multiple=True,
)
@click.option(
    "--file-friendly-logging",
    is_flag=True,
    help="Outputs progress bar status on separate lines and slows refresh rate",
)
def run(
    experiment: str,
    directory: Optional[Union[str, os.PathLike]] = None,
    overrides: Optional[str] = None,
    include_package: Optional[List[str]] = None,
    file_friendly_logging: bool = False,
):
    """
    Run a tango experiment.
    """
    _run(
        experiment,
        directory=directory,
        overrides=overrides,
        include_package=include_package,
        file_friendly_logging=file_friendly_logging,
    )


def _run(
    experiment: str,
    directory: Optional[Union[str, os.PathLike]] = None,
    overrides: Optional[str] = None,
    include_package: Optional[List[str]] = None,
    file_friendly_logging: bool = False,
):
    if file_friendly_logging:
        os.environ["FILE_FRIENDLY_LOGGING"] = "true"

    from pathlib import Path
    from tango.common.file_lock import FileLock
    from tango.common.params import Params
    from tango.common.util import import_module_and_submodules
    from tango.executor import Executor
    from tango.step_cache import StepCache
    from tango.step_graph import StepGraph

    logger = logging.getLogger("tango")

    # Import included packages to find registered components.
    if include_package is not None:
        for package_name in include_package:
            import_module_and_submodules(package_name)

    # Read params.
    params = Params.from_file(experiment, params_overrides=overrides or "")

    # Prepare directory.
    if directory is None:
        from tempfile import mkdtemp

        directory = mkdtemp(prefix="tango-")
        click.echo(
            "Creating temporary directory for run: " + click.style(f"{directory}", fg="yellow")
        )
    directory = Path(directory)
    directory.mkdir(parents=True, exist_ok=True)

    # Initialize step graph, executor, and cache.
    step_graph = StepGraph(params.pop("steps", keep_as_dict=True))
    executor = Executor.from_params(
        params.pop("executor", default={}), include_package=include_package
    )
    step_cache = StepCache.from_params(
        params.pop("cache", default={}), dir=directory / "step_cache"
    )

    # Acquire lock on directory.
    directory_lock = FileLock(directory / ".lock", read_only_ok=True)
    directory_lock.acquire_with_updates(desc="acquiring directory lock...")

    try:
        # Remove symlinks to old results.
        for filename in directory.glob("*"):
            if filename.is_symlink():
                relative_target = os.readlink(filename)
                if not relative_target.startswith("step_cache/"):
                    continue
                logger.debug(
                    f"Removing symlink '{filename.name}' to previous result {relative_target}"
                )
                filename.unlink()

        # Produce results and symlink everything that has been computed.
        for step in executor.execute_step_graph(step_graph, step_cache):
            name = step.name
            if step in step_cache:
                step_link = directory / name
                if step_link.exists():
                    step_link.unlink()
                step_link.symlink_to(
                    step_cache.path_for_step(step).relative_to(directory),
                    target_is_directory=True,
                )
                click.echo(
                    click.style("✓ The output for ", fg="green")
                    + click.style(f'"{name}"', bold=True, fg="green")
                    + click.style(" is in ", fg="green")
                    + click.style(f"{step_link}.", bold=True, fg="green")
                )
    finally:
        directory_lock.release()


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Generic, TypeVar

import dill
import torch

from tango.format import Format
from tango.common.util import PathOrStr


T = TypeVar("T")


@Format.register("torch")
class TorchFormat(Format[T], Generic[T]):
    """
    This format writes the artifact using ``torch.save()``.

    Unlike :class:`tango.format.DillFormat`, this has no special support for iterators.

    .. tip::

        Registered as a :class:`~tango.format.Format` under the name "torch".

    """

    VERSION = 2

    def write(self, artifact: T, dir: PathOrStr):
        filename = Path(dir) / "data.pt"
        with open(filename, "wb") as f:
            torch.save((self.VERSION, artifact), f, pickle_module=dill)

    def read(self, dir: PathOrStr) -> T:
        filename = Path(dir) / "data.pt"
        with open(filename, "rb") as f:
            version, artifact = torch.load(f, pickle_module=dill, map_location=torch.device("cpu"))
            if version > self.VERSION:
                raise ValueError(
                    f"File {filename} is too recent for this version of {self.__class__}."
                )
            return artifact
